"""
Download the Img Gen model from Hugging Face
"""
import os
import click


'''set the model download cache directory'''

from dataclasses import dataclass
@dataclass
class DirectorySetting:
    """set the directory for the model download"""
    home_dir: str="/home/jovyan/llm-models"
    transformers_cache_home: str="core-kind/yinwang/models"
    huggingface_token_file: str="core-kind/yinwang/.cache/huggingface/token"

    def get_cache_home(self):
        """get the cache home"""
        return f"{self.home_dir}/{self.transformers_cache_home}"
    
    def get_token_file(self):
        """get the token file"""
        return f"{self.home_dir}/{self.huggingface_token_file}"

# https://huggingface.co/runwayml/stable-diffusion-v1-5
model_map = {
    "diffusion-v1.5":     "runwayml/stable-diffusion-v1-5",
}

dir_mode_map = {
    "kf_notebook": DirectorySetting(),
    "mac_local": DirectorySetting(home_dir="/Users/yingding", transformers_cache_home="MODELS", huggingface_token_file="MODELS/.huggingface_token"),
}

default_model_type = "diffusion-v1.5"
default_dir_mode = "kf_notebook"


def need_token(model_type: str, model_name_prefix: list=["llama", "gemma"]):
    """check if the model needs token"""
    return any([model_type.startswith(prefix) for prefix in model_name_prefix])


def get_token(dir_setting: DirectorySetting):
    """get the token from the token file"""
    token_file_path = dir_setting.get_token_file()
    with open(token_file_path, "r") as file:
        # file read add a new line to the token, remove it.
        token = file.read().replace('\n', '')
    return token

# print the raw string to see if there is new line in the token
# print(r'{}'.format(token))

# Reference: https://click.palletsprojects.com/en/8.1.x/quickstart/
# @click.option(..., is_flag=True, ...) set the option to be boolean
# call with download_llms --help
# https://www.geeksforgeeks.org/argparse-vs-docopt-vs-click-comparing-python-command-line-parsing-libraries/
# https://click.palletsprojects.com/en/8.1.x/quickstart/
# https://click.palletsprojects.com/en/8.1.x/options/
# https://www.youtube.com/watch?v=kNke39OZ2k0
@click.command()
@click.option('-t','--model-type', 'model_type', default=default_model_type, type=str, required=False, help=f"set the llm type to download:\n{', '.join(model_map.keys())}, default is {default_model_type}")
@click.option('-m','--mode', 'dir_mode', default=default_dir_mode, type=str, required=False, help=f"set the directory settings to use:\n{', '.join(dir_mode_map.keys())}, default is {default_dir_mode}")
def download(model_type: str=default_model_type, dir_mode: str=default_dir_mode):
    """
    This method will download the llm model. If cache exists, the cached model will be used.
    
    get help:\n
    python3 download_img_gen.py --help

    valid call:\n
    python3 download_img_gen.py -t diffusion-v1.5\n
    python3 download_img_gen.py --model-type diffusion-v1.5\n
    
    invalid call:\n
    python3 download_img_gen.py -t=diffusion-v1.5\n
    python3 download_img_gen.py --model-type=diffusion-v1.5\n

    set directory:\n
    python3 download_img_gen.py -t diffusion-v1.5 -m kf_notebook
    
    Args:\n
      model_type: "diffusion-v1.5"\n
      dir_mode: "kf_notebook", "mac_local"\n
    """
 
    dir_setting=dir_mode_map.get(dir_mode, dir_mode_map[default_dir_mode])
    os.environ['XDG_CACHE_HOME']=dir_setting.get_cache_home()
    print("-"*10)
    print(f"download dir: {os.environ['XDG_CACHE_HOME']}")

    from diffusers import StableDiffusionPipeline
    import torch

    model_name = model_map.get(model_type, model_map[default_model_type])
    
    print("-"*10)
    print(f"model_type: {model_type}")
    print(f"model_name: {model_name}")
    print("-"*10)
    
    if need_token(model_type):
        # kwargs = {"use_auth_token": get_token(dir_setting)}
        kwargs = {
            "torch_dtype": torch.float16,
            "token": get_token(dir_setting)
        }
        print("huggingface token loaded")
    else:
        kwargs = {
            "torch_dtype": torch.float16
        }
        print("huggingface token is NOT needed")
    print("-"*10)

    model = StableDiffusionPipeline.from_pretrained(model_name, **kwargs)
    
if __name__ == '__main__':
    """
    uses default for click
    https://stackoverflow.com/questions/49011223/python-correct-use-of-click-with-main-and-init
    """
    download()


def solution(lst):
	"""Given a non-empty list of integers, return the sum of all of the odd elements that are in even poitions.
    
	Examples
    solution([5, 8, 7, 1]) = 12
    solution([3, 3, 3, 3, 3]) = 9
    solution([30, 13, 24, 321]) = 0
    """
	
    

print(solution([5, 8, 7, 1]))
print(solution([3, 3, 3, 3, 3]))
print(solution([30, 13, 24, 321]))	


from dataclasses import dataclass
import sqlalchemy, os

# import psycopg


class SqlDBConfig:
    def __init__(
        self,
        host: str = "localhost",
        user: str = "postgres",
        password: str = "postgres",
        dbname: str = "postgres",
        schema: str = "public",
        port: int = 5432,
    ):
        self.DB_HOST = host
        self.DB_USER = user
        self.DB_USER_PW = password
        self.DB_NAME = dbname
        self.DB_SCHEMA = schema
        self.DB_PORT = port

    @classmethod
    def load_env(clz, key: str, default=""):
        return os.environ.get(key, default)

    def __repr__(self) -> str:
        """for program and developer obj representation also used by jupyter notebook cell"""
        return f"{self.DB_HOST}\n{self.DB_USER}\n{self.DB_USER_PW}\n{self.DB_NAME}\n{self.DB_SCHEMA}\n{self.DB_PORT}"

    def __str__(self):
        """for user print() function"""
        return self.__repr__()

    def to_psycopg_dic(self) -> dict:
        return {
            "dbname": self.DB_NAME,
            "user": self.DB_USER,
            "password": self.DB_USER_PW,
            "host": self.DB_HOST,
            "port": self.DB_PORT,
        }

    def to_sqlalchemy_url(self) -> sqlalchemy.engine.url.URL:
        return sqlalchemy.URL.create(
            drivername="postgresql+psycopg",
            username=self.DB_USER,
            password=self.DB_USER_PW,  # plain (unescaped) text
            host=self.DB_HOST,
            database=self.DB_NAME,
            port=self.DB_PORT,
        )


class SqlDBHelperFactory:
    """Factory class for creating sqlalchemy engine"""

    @staticmethod
    def get_sync_engine(
        db_config: SqlDBConfig, verbose: bool = False
    ) -> sqlalchemy.engine.base.Engine:
        # create sync engine https://docs.sqlalchemy.org/en/20/core/engines.html
        return sqlalchemy.create_engine(db_config.to_sqlalchemy_url(), echo=verbose)

    @staticmethod
    def get_db_config_from_env(
        host_key: str = "SCIVIAS_DB_HOST",
        user_key: str = "SCIVIAS_DB_USERNAME",
        password_key: str = "SCIVIAS_DB_PASSWORD",
        db_key: str = "SCIVIAS_DB_NAME",
        schema_key: str = "SCIVIAS_DB_SCHEMA",
        port_key: str = "SCIVIAS_DB_PORT",
    ) -> SqlDBConfig:
        return SqlDBConfig(
            host=SqlDBConfig.load_env(host_key),
            user=SqlDBConfig.load_env(user_key),
            password=SqlDBConfig.load_env(password_key),
            dbname=SqlDBConfig.load_env(db_key),
            schema=SqlDBConfig.load_env(schema_key),
            port=int(SqlDBConfig.load_env(port_key, "5432")),
        )



import re

def parse_response_str(response: str) -> str:
    # The (?s) is a mode modifier for regular expressions in Python. It's also known as a flag. 
    # This specific flag, (?s), is short for DOTALL, which is a mode that allows . to match any character 
    # including newline characters.
    # the dot . matches any character except a newline. You can use the DOTALL mode to match new line.
    # replace // patient weight in kilograms\n} with \n}
    # return re.sub(r"(?s)//.*}", "}", response) is a short form of the following
    return re.sub(r"//.*}", "\n}", response, flags=re.DOTALL)

def search_response_str(response: str) -> str:
    if not re.search(r"}.*```", response, flags=re.DOTALL):
        post_proccessed_response = response + "}\n```"
    else:
        post_proccessed_response = response

    # remove the leading and tailing text outside the ``` ```
    post_proccessed_response = re.search(
        r"```[\s\S]+```", post_proccessed_response
    ).group(0)

    # remove the comment // patient weight in kilogram \n} and keep the \n}
    # re.DOTALL is a flag that makes . match newlines as well, since default . doesn't match newline characters
    post_proccessed_response = re.sub(r"//.*}", "\n}", post_proccessed_response, flags=re.DOTALL)
    return post_proccessed_response


def exp_1():
    print("exp_1")
    response_1 = '```json { "patient_weight": "43.2" // patient weight in kilogram }'
    response_2 = '```json { "patient_weight": "45.2" }'
    response_3 = '```json\n{\n\t"patient_weight": "43.2"  // patient weight in kilograms\n}\n'

    responses = [response_1, response_2, response_3]    
    for response in responses:
        processed_response = parse_response_str(response)
        print(repr(processed_response))  # json { "patient_weight": "43.2" }


def exp_2():
    print("exp_2")
    response_1 = '```json { "patient_weight": "43.2"'
    response_2 = '```json { "patient_weight": "45.2" '
    response_3 = '```json\n{\n\t"patient_weight": "43.2" \n'
    response_4 = '```json { "patient_weight": "45.2" }\n```'
    response_5 = '```json { "patient_weight": "45.2"\n}\n```'
    response_6 = '```json { "patient_weight": "45.2" }```'

    responses = [response_1, response_2, response_3, response_4, response_5, response_6 ]    
    for response in responses:
        processed_response = search_response_str(response)
        print(repr(processed_response))

if __name__ == '__main__':
    exp_1()
    exp_2()


from .accelerator_utils import (
    DirectorySetting, DIR_MODE_MAP, TokenHelper, AcceleratorHelper,
    AcceleratorStatus, MpsAcceleratorStatus, CudaAcceleratorStatus,
)

__all__ = [ DirectorySetting, DIR_MODE_MAP, TokenHelper, AcceleratorHelper,
            AcceleratorStatus, MpsAcceleratorStatus, CudaAcceleratorStatus,]


"""
Before run this script, need to generate a token from huggingface
To login, `huggingface_hub` requires a token generated from https://huggingface.co/settings/tokens
-> user profile -> settings -> Access Tokens -> New token
name: llama2
role: read

Then call
huggingface-cli

use 
echo "<my_token>" > token
to create a token instead of using vi, which will add a new line char to token automatically.

"""
import os
import click
# import argpass


'''set the model download cache directory'''
# set the home directory for the models

from dataclasses import dataclass
@dataclass
class DirectorySetting:
    """set the directory for the model download"""
    home_dir: str="/home/jovyan/llm-models"
    transformers_cache_home: str="core-kind/yinwang/models"
    huggingface_token_file: str="core-kind/yinwang/.cache/huggingface/token"

    def get_cache_home(self):
        """get the cache home"""
        return f"{self.home_dir}/{self.transformers_cache_home}"
    
    def get_token_file(self):
        """get the token file"""
        return f"{self.home_dir}/{self.huggingface_token_file}"

model_map = {
    "llama7B-chat":     "meta-llama/Llama-2-7b-chat-hf",
    "llama13B-chat":    "meta-llama/Llama-2-13b-chat-hf",
    "llama70B-chat":    "meta-llama/Llama-2-70b-chat-hf",
    "llama3-8B-inst":   "meta-llama/Meta-Llama-3-8B-Instruct",
    "llama3-70B-inst":  "meta-llama/Meta-Llama-3-70B-Instruct",
    "mistral7B-01":     "mistralai/Mistral-7B-v0.1",
    "mistral7B-inst02": "mistralai/Mistral-7B-Instruct-v0.2",
    "mixtral8x7B-01":   "mistralai/Mixtral-8x7B-v0.1",
    "mixtral8x7B-inst01": "mistralai/Mixtral-8x7B-Instruct-v0.1", 
    "gemma7b-it": "google/gemma-7b-it",
    "gemma7b":    "google/gemma-7b",
    "gemma2b-it": "google/gemma-2b-it",
    "gemma2b":    "google/gemma-2b",
    "gemma7b-it-1.1": "google/gemma-1.1-7b-it",
    "gemma2b-it-1.1": "google/gemma-1.1-2b-it",
    "phi3-medium-128k-inst": "microsoft/Phi-3-medium-128k-instruct",
}

dir_mode_map = {
    "kf_notebook": DirectorySetting(),
    "mac_local": DirectorySetting(home_dir="/Users/yingding", transformers_cache_home="MODELS", huggingface_token_file="MODELS/.huggingface_token"),
}

default_model_type = "mistral7B-01"
default_dir_mode = "kf_notebook"


def need_token(model_type: str, model_name_prefix: list=["llama", "gemma"]):
    """check if the model needs token"""
    return any([model_type.startswith(prefix) for prefix in model_name_prefix])


def get_token(dir_setting: DirectorySetting):
    """get the token from the token file"""
    token_file_path = dir_setting.get_token_file()
    with open(token_file_path, "r") as file:
        # file read add a new line to the token, remove it.
        token = file.read().replace('\n', '')
    return token

# print the raw string to see if there is new line in the token
# print(r'{}'.format(token))

# Reference: https://click.palletsprojects.com/en/8.1.x/quickstart/
# @click.option(..., is_flag=True, ...) set the option to be boolean
# call with download_llms --help
# https://www.geeksforgeeks.org/argparse-vs-docopt-vs-click-comparing-python-command-line-parsing-libraries/
# https://click.palletsprojects.com/en/8.1.x/quickstart/
# https://click.palletsprojects.com/en/8.1.x/options/
# https://www.youtube.com/watch?v=kNke39OZ2k0
@click.command()
@click.option('-t','--model-type', 'model_type', default=default_model_type, type=str, required=False, help=f"set the llm type to download:\n{', '.join(model_map.keys())}, default is {default_model_type}")
@click.option('-m','--mode', 'dir_mode', default=default_dir_mode, type=str, required=False, help=f"set the directory settings to use:\n{', '.join(dir_mode_map.keys())}, default is {default_dir_mode}")
def download(model_type: str=default_model_type, dir_mode: str=default_dir_mode):
    """
    This method will download the llm model. If cache exists, the cached model will be used.
    
    get help:
    python3 download_llms.py --help

    valid call:
    python3 download_llms.py -t mistral7B-01
    python3 download_llms.py --model-type mistral7B-01
    
    invalid call:
    python3 download_llms.py -t=mistral7B-01
    python3 download_llms.py --model-type=mistral7B-01

    set directory:
    python3 download_llms.py -t mistral7B-01 -m kf_notebook
    
    Args:
      model_type: "llama7B-chat", "llama13B-chat", "llama70B-chat", "mistral7B-01", "mistral7B-inst02", "mistral8x7B-01"
      dir_mode: "kf_notebook", "mac_local"
    """
 
    dir_setting=dir_mode_map.get(dir_mode, dir_mode_map[default_dir_mode])
    os.environ['XDG_CACHE_HOME']=dir_setting.get_cache_home()
    print("-"*10)
    print(f"download dir: {os.environ['XDG_CACHE_HOME']}")

    from transformers import pipeline, AutoModelForCausalLM, AutoTokenizer

    model_name = model_map.get(model_type, model_map[default_model_type])
    
    print("-"*10)
    print(f"model_type: {model_type}")
    print(f"model_name: {model_name}")
    print("-"*10)
    
    kwargs = {"trust_remote_code" : True}
    if need_token(model_type):
        # kwargs = {"use_auth_token": get_token(dir_setting)}
        kwargs = {
            **kwargs,
            "token": get_token(dir_setting)
        }
        print("huggingface token loaded")
    else:
        print("huggingface token is NOT needed")
    print("-"*10)
    tokenizer = AutoTokenizer.from_pretrained(model_name, **kwargs)
    model = AutoModelForCausalLM.from_pretrained(model_name, **kwargs)
    
    # all data will be saved in the local_dir, is not what we want
    # local_dir=dir_setting.get_cache_home()
    # print(f"local_dir: {local_dir}")

    # from huggingface_hub import snapshot_download, hf_hub_download
    # # download all files https://github.com/ggerganov/llama.cpp/discussions/2948
    # snapshot_download(repo_id=model_name, local_dir=local_dir,
    #                  local_dir_use_symlinks=False, revision="main")
    
if __name__ == '__main__':
    """
    uses default for click
    https://stackoverflow.com/questions/49011223/python-correct-use-of-click-with-main-and-init
    """
#    parser = argparse.ArgumentParser()
#    parser.add_argument('--model-type', dest='model-type',
#                        default=os.environ['model-type'], type=str, help='the llama2 type to download: 7B, 13B')
    
#    args = parser.parse_args()
#    params_dict = args.__dict__
    
#    print(params_dict)
    download()


from .builder import (
    KwargsBuilder, PromptHelper, ModelInfo, ModelCatalog
    )
from .config import (LocalCausalLMConfig, ModelConfig)
from .components import (StructuredOutputParserHelper)


__all__ = [KwargsBuilder, PromptHelper, ModelInfo, ModelCatalog, ModelConfig, LocalCausalLMConfig, StructuredOutputParserHelper]


import torch
import os
import subprocess
import re
import sys
import time
from dataclasses import dataclass
import Warnings


@dataclass
class DirectorySetting:
    """set the directory for the model download"""

    home_dir: str = "/home/jovyan/llm-models"
    transformers_cache_home: str = "core-kind/yinwang/models"
    huggingface_token_file: str = "core-kind/yinwang/.cache/huggingface/token"

    def get_cache_home(self):
        """get the cache home"""
        return os.path.join(f"{self.home_dir}", f"{self.transformers_cache_home}")

    def get_token_file(self):
        """get the token file"""
        # use path join for windows and linux compatibility
        return os.path.join(f"{self.home_dir}", f"{self.huggingface_token_file}")


DIR_MODE_MAP = {
    "kf_notebook": DirectorySetting(),
    "mac_local": DirectorySetting(
        home_dir="/Users/yingding",
        transformers_cache_home="MODELS",
        huggingface_token_file="MODELS/.huggingface_token",
    ),
    "win_local": DirectorySetting(
        home_dir="C:\\Users\\yingdingwang",
        transformers_cache_home="MODELS",
        huggingface_token_file="MODELS\\.huggingface_token",
    ),
}


class TokenHelper:
    def __init__(self, dir_setting: DirectorySetting, prefix_list: list = ["llama"]):
        self.prefix_list = prefix_list
        self.dir_setting = dir_setting

    def need_token(self, model_type: str) -> bool:
        """check if the model needs token"""
        # check if the model type startswith a string in the prefix list
        # model_type.startswith(model_name_prefix)
        return any([model_type.startswith(prefix) for prefix in self.prefix_list])

    def get_token(self):
        """get the token from the token file"""
        token_file_path = self.dir_setting.get_token_file()
        with open(token_file_path, "r") as file:
            # file read add a new line to the token, remove it.
            token = file.read().replace("\n", "")
        return token

    def gen_token_kwargs(self, model_type: str):
        """ """
        if self.need_token(model_type):
            # kwargs = {"use_auth_token": get_token(dir_setting)}
            token_kwargs = {
                "token": self.get_token(),
                # "truncation_side": "left",
                # "return_tensors": "pt",
            }
            print("huggingface token loaded")
        else:
            token_kwargs = {}
            print("huggingface token is NOT needed")
        return token_kwargs


class AcceleratorHelper:
    @staticmethod
    def print_container_info() -> None:
        print("-" * 10)
        print(time.strftime("%Y-%m-%d_%H-%M-%S"))
        print(f"python version: {sys.version}")
        print(f"torch version: {torch.__version__}")
        print("-" * 10)

    @staticmethod
    def nvidia_device_info() -> str:
        """get the nvidia MIGs device uuid and GPU uuid"""
        # blocking call
        result = subprocess.run(["nvidia-smi", "-L"], stdout=subprocess.PIPE)
        # decode the byte object, returns string with \n
        cmd_out_str = result.stdout.decode("utf-8")
        return [line.strip() for line in cmd_out_str.split("\n") if len(line) > 0]

    @staticmethod
    def extract_nvidia_device_uuids(input: str):
        """parse the nvidia devices uuid from the nvidia device info str"""
        try:
            # r'' before the search pattern indicates it is a raw string,
            # otherwise "" instead of single quote
            uuid = re.search(r"UUID\:\s(.+?)\)", input).group(1)
        except AttributeError:
            # "UUID\:\s" and "\)" not found
            uuid = ""
        return uuid

    @staticmethod
    def nvidia_device_uuids_filtered_by(
        is_mig: bool = False, log_output: bool = False
    ) -> str:
        """get a comma separated str of nvidia MIGs devices"""
        info_list = AcceleratorHelper.nvidia_device_info()
        if is_mig:
            # skip the first GPU ID, get the MIGs IDS
            uuid_list = [
                AcceleratorHelper.extract_nvidia_device_uuids(e) for e in info_list[1:]
            ]
        else:  # all GPU devices
            uuid_list = [
                AcceleratorHelper.extract_nvidia_device_uuids(e) for e in info_list
            ]
        if log_output is not None and log_output:
            print(uuid_list)

        # if multi gpus need to join the device together for pytorch
        return ",".join(uuid_list)

    @staticmethod
    def _init_cuda_torch(
        uuids: str, dir_setting: DirectorySetting, debug: bool = False
    ) -> None:
        """setup the default env variables for transformers

        Args:
          uuids: a comma separate str of nvidia gpu/mig uuids
        """
        os.environ["WORLD_SIZE"] = "1"
        os.environ["CUDA_VISIBLE_DEVICES"] = uuids
        os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:512"  # 512
        os.environ["XDG_CACHE_HOME"] = dir_setting.get_cache_home()
        # os.environ['XDG_CACHE_HOME'] = f"{data_path}/models"
        if debug:
            # for debugging
            os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
        

    # @staticmethod
    # def init_mps_torch(dir_setting: DirectorySetting) -> None:
    #     """setup the default env variables for transformers"""
    #     # os.environ['TOKENIZERS_PARALLELISM'] = "false"
    #     # https://github.com/huggingface/transformers/issues/5486?ref=assemblyai.com
    #     os.environ["WORLD_SIZE"] = "1"
    #     os.environ["XDG_CACHE_HOME"] = dir_setting.get_cache_home()

    @staticmethod
    def init_torch_env(
        accelerator: str, dir_setting: DirectorySetting, **kwargs
    ) -> None:
        """
        Initializes the PyTorch environment based on the specified accelerator type.

        Args:
            accelerator (str): The type of accelerator to use. Supported values are "mps", "xpu", "npu", "cpu" and "cuda".
            dir_setting (DirectorySetting): An instance of DirectorySetting to manage directory paths.
            kwargs (dict): Additional keyword arguments. For "cuda" accelerator, it can include:
                - "uuids" (str): A string representing the UUIDs.
                - "debug" (bool, optional): A flag to enable or disable debug mode. Defaults to False.

        Raises:
            Warnings.warn: If the specified accelerator is not supported.
        """
        if (
            accelerator == "mps"
            or accelerator == "xpu"
            or accelerator == "npu"
            or accelerator == "cpu"
        ):
            os.environ["WORLD_SIZE"] = "1"
            os.environ["XDG_CACHE_HOME"] = dir_setting.get_cache_home()
        elif accelerator == "cuda":
            if "uuids" in kwargs and isinstance(kwargs["uuids"], str):
                uuids = kwargs["uuids"]
                if "debug" in kwargs and isinstance(kwargs["debug"], bool):
                    debug = kwargs["debug"]
                else:
                    debug = False
                AcceleratorHelper._init_cuda_torch(
                    uuids=uuids, dir_setting=dir_setting, debug=debug
                )
            else:
                # warning with Warnings class
                Warnings.warn("UUIDs not provided for CUDA accelerator")
        else:
            # warning with Warnings class
            Warnings.warn("accelerator not supported")


class AcceleratorStatus:
    # Reference: https://stackoverflow.com/questions/58216000/get-total-amount-of-free-gpu-memory-and-available-using-pytorch
    # from typing import Tuple
    def byte_gb_info(self, byte_mem) -> str:
        """calculate the byte size to GB size for better human readable"""
        # format the f string float with :.2f to decimal digits
        # https://zetcode.com/python/fstring/
        return f"{(byte_mem/1024**3):4f} GB"

    def gpu_usage(self) -> None:
        pass

    # factory method to create the correct accelerator status class
    @staticmethod
    def create_accelerator_status() -> "AcceleratorStatus":
        # check if the MPS is enabled
        if torch.backends.mps.is_available():
            return MpsAcceleratorStatus()
        elif torch.cuda.is_available():
            return CudaAcceleratorStatus()
        else:
            return AcceleratorStatus()


class MpsAcceleratorStatus(AcceleratorStatus):
    def __init__(self):
        pass

    def accelerator_mem_info(self):
        # allocated
        a = torch.mps.driver_allocated_memory()
        print(f"Allocated memory : {self.byte_gb_info(a)}")

    def gpu_usage(self) -> None:
        print("-" * 20)
        self.accelerator_mem_info()
        print("-" * 20)


class CudaAcceleratorStatus(AcceleratorStatus):
    def __init__(self):
        pass

    def accelerator_mem_info(self, device_idx: int):
        # total
        t = torch.cuda.get_device_properties(device_idx).total_memory
        # usable
        r = torch.cuda.memory_reserved(device_idx)
        # allocated
        a = torch.cuda.memory_allocated(device_idx)
        # still free
        f = r - a
        # unit = "GB"
        print(  # "GPU memory info:\n" +
            f"Physical  memory : {self.byte_gb_info(t)}\n"
            + f"Reserved  memory : {self.byte_gb_info(r)}\n"
            + f"Allocated memory : {self.byte_gb_info(a)}\n"
            + f"Free      memory : {self.byte_gb_info(f)}"
        )

    def accelerator_compute_info(self, device_idx: int) -> None:
        name = torch.cuda.get_device_properties(device_idx).name
        count = torch.cuda.get_device_properties(device_idx).multi_processor_count
        print(
            f"Device name      : {name} \n"
            + f"Device idx       : {device_idx} \n"
            + f"No. of processors: {count}"
        )

    def gpu_usage(self) -> None:
        num_of_gpus = torch.cuda.device_count()
        # this shows only the gpu device, not the MIG
        print(f"num_of_gpus: {num_of_gpus}")
        for device_idx in range(torch.cuda.device_count()):
            print("-" * 20)
            self.accelerator_compute_info(device_idx)
            self.accelerator_mem_info(device_idx)
            print("-" * 20)
